'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var update = _interopDefault(require('immutability-helper'));

function tail(arr, offset) {
  if (offset === void 0) {
    offset = 0;
  }

  if (!arr || !arr.length) {
    return undefined;
  }

  if (offset < 0) {
    return arr[arr.length - 1];
  }

  return arr[Math.max(0, arr.length - offset - 1)];
}
function resolvePathKeys(path) {
  if (path == null || !path.toString) {
    return [];
  }

  if (Array.isArray(path)) {
    return path.length ? path.map(function (seg) {
      return seg.toString();
    }) : [];
  }

  return path === "" ? [] : path.toString().split(".");
}
function createTreeList(pathKeys) {
  var target = {};
  var treeList = [target];

  for (var i = 0; i < pathKeys.length; i++) {
    var key = pathKeys[i];
    target[key] = {};
    target = target[key];
    treeList.push(target);
  }

  return treeList;
}
function toUpdateQuery(cmd, path, payload) {
  var pathKeys = resolvePathKeys(path);
  var treeList = createTreeList(pathKeys);

  switch (cmd) {
    case "$push":
    case "$unshift":
    case "$add":
    case "$remove":
      tail(treeList)[cmd] = Array.isArray(payload) ? payload : [payload];
      break;

    case "$splice":
      tail(treeList)[cmd] = Array.isArray(payload[0]) ? payload : [payload];
      break;

    case "$set":
    case "$merge":
    case "$apply":
      tail(treeList)[cmd] = payload;
      break;

    case "$toggle":
    case "$unset":
      if (!pathKeys.length) {
        return {};
      }

      if (payload) {
        tail(treeList)[cmd] = Array.isArray(payload) ? payload : [payload];
      } else {
        var target = tail(treeList, 1);
        var key = tail(pathKeys);
        delete target[key];
        target[cmd] = [key];
      }

      break;
  }

  return treeList[0];
}

var ioc = {
  update: update,
  toUpdateQuery: toUpdateQuery
};

function resolveFirstOptional(args, count, defaultValue) {
  return args.length < count ? [defaultValue].concat(args) : args;
}

function createCommandFunction(cmd, destIsOptional) {
  return function (src) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    var _ref = destIsOptional ? resolveFirstOptional(args, 2, "") : args,
        dest = _ref[0],
        payload = _ref[1];

    var query = ioc.toUpdateQuery(cmd, dest, payload);
    return ioc.update(src, query);
  };
}

var $update = update;
var $push = createCommandFunction("$push", true);
var $unshift = createCommandFunction("$unshift", true);
var $splice = createCommandFunction("$splice", true);
var $set = createCommandFunction("$set", false);
var $toggle = createCommandFunction("$toggle", false);
var $unset = createCommandFunction("$unset", false);
var $merge = createCommandFunction("$merge", true);
var $apply = createCommandFunction("$apply", true);
var $add = createCommandFunction("$add", true);
var $remove = createCommandFunction("$remove", true);

exports.$update = $update;
exports.$push = $push;
exports.$unshift = $unshift;
exports.$splice = $splice;
exports.$set = $set;
exports.$toggle = $toggle;
exports.$unset = $unset;
exports.$merge = $merge;
exports.$apply = $apply;
exports.$add = $add;
exports.$remove = $remove;
